# namespace eval Json {
    # variable reMaskJsonChars "\[\\b\\t\\n\\f\\r\\v\\x22\\x5c\]" ; # BS TAB LF FF CR " \\
	# variable maskJsonChars [dict create 8 "\\b" 9 "\\t" 10 "\\n" 11 "\\v" 12 "\\f" 13 "\\r" 34 "\\\"" 92 "\\\\"]
# }
proc getDictValue {dict key default} {	# returns value from dictionary with possible default
	if [dict exists $dict $key] {
		return [dict get $dict $key]
	}
	return $default
}
proc fg_Restore {} {
	global Elem
	global defaultsMap
	fg_eventsHandler [list "Reset"]
	foreach id [dict keys $Elem] {
		setValue $id [getDictValue $defaultsMap $id ""]
	}
}
proc valueOf {name} {
	global Elem wReference fieldsTypeMap keyValueMap defaultsMap QSMap scaleMap
	set wdg [getDictValue $wReference $name ""]
	set value ""
	set type [getDictValue $fieldsTypeMap $name "T"]
	if {$type == "H"} {
		set value [getDictValue $defaultsMap "$name" $value]
	}
	if {$wdg != ""} {
			switch $type {
			A {
				set value [$wdg get 1.0 end]
			}
			B {}
			CMB - CMT {
				set value [$wdg get]
				set value [getDictValue $keyValueMap "$name$value" $value]
			}
			CKB {
				set globalName "::$name"
				set value [expr \$$globalName]	; # check box has an associated name
			}
			S - QS {
				set value [$wdg get]
				if {$type == "QS"} {
					set lst [dict get $QSMap $name]
					set value [lindex $lst [expr int($value/10)]]
				} else {
					set decimals [dict get $scaleMap $name]
					if {$decimals < 1} {set value [expr {int($value/pow(10,abs($decimals)))}]
					} else {
						set decimals [expr {pow(10,$decimals)}]
						set value [expr {int($value*$decimals)/$decimals}]
					}
				}
			}
			R {
				set globalName "::$name"
				set value [expr \$$globalName]	; # RDB has an associated name
			}
			default {
				set value [$wdg get]
			}
		}
	}
	return $value	; 	# also hidden fields
}
proc setValue {name value} {
	global wReference fieldsTypeMap rdbMap QSMap
	set wName [string tolower $name]
	set wdg [getDictValue $wReference $name ""]
	if {$wdg != ""} {
		set type [getDictValue $fieldsTypeMap $name "T"]
		switch $type {
			B {}
			A {
				$wdg delete 1.0 end
				$wdg insert 1.0 $value
			}
			CMB - CMT {
				$wdg set $value
			}
			CKB {
				set globalName "::$name"
				set $globalName $value	; # check box has an associated name
			}
			F - WF - D {
				set wdg ".fg.f[string tolower $name]"
				set v [$wdg get]
				$wdg delete 0 [string length $v]
				$wdg insert 0 $value

			}
			R {
				set globalName "::$name"
				set $globalName $value	; # RDB has an associated name
				if {$value != "" } {[dict get $rdbMap "$name$value"] invoke}
			}
			QS - S {
				if {$value != ""} {
					set fg::fgls_$wName $value
					if {$type == "QS"} {
						set lst [dict get $QSMap $name]
						set indx [lsearch -exact $lst $value]
						set value [expr $indx * 10 + 9];	# like ($index+1)*10-1
					}
					$wdg configure -value $value
				}
			}
			default {	# text fields
				set v [$wdg get]
				$wdg delete 0 [string length $v]
				$wdg insert 0 $value
			}
		}
	}
}
proc splitKeyValue {name items} {
	global keyValueMap
	set lst [split $items "|"]
	set l []
	foreach x $lst {
		append x ":" $x
		set keyValue [split $x ":"]
		set value [lindex $keyValue 1]
		dict set keyValueMap "$name$value" [lindex $keyValue 0]
		lappend l [lindex $keyValue 1]
	}
	return $l
}
proc fg_setCombo {name items} {
	global keyValueMap
	set keyValueMap [dict remove $keyValueMap {*}[dict keys $keyValueMap "$name*"]] ; #	remove old keys
#	lappend lst {*}[splitKeyValue $name "|$items"]
	.fg.[string tolower $name] configure -values [splitKeyValue $name "|$items"]
	
}
proc updateControls {data} {
	global controlsMap
	lassign $data k alert
	regexp -nocase {^s*(\w+)\s+(is)\s+(.*)} $k -> key operator operand
	set operator [string tolower $operator]
	if {[dict exists $controlsMap $key]} {
		set a [dict get $controlsMap $key]
		set lst "$operator $operand [list $alert]"
		lappend a $lst
		dict set controlsMap $key $a
	} else {
		set lst "$operator $operand [list $alert]"
		dict set controlsMap $key [list $lst]
	}
}
proc controlField {name answer} {
	global controlsMap fieldsTypeMap
	set ctrls [getDictValue $controlsMap $name ""]
	if {$ctrls == ""} {return ""}
	set sErrors ""
	set wName [string tolower $name]
	.fg.fgl_$wName configure -foreground black
	set type [getDictValue $fieldsTypeMap $name "T"]
	foreach ctrl $ctrls {
		lassign $ctrl operator operand alert
		if {$operator == "is"} {
			switch [string tolower $operand] {
				mail {
					if {![regexp "^\[a-zA-Z0-9._-\]+@\[a-zA-Z0-9.-\]+\\.\[a-zA-Z\]{2,4}$|^$" $answer]} {
						.fg.fgl_$wName configure -foreground red
						append sErrors "\n$name:\t$alert"
					}
				}
				required {
					if {[regexp "^\\s*$" $answer]} {
						.fg.fgl_$wName configure -foreground red
						append sErrors "\n$name:\t$alert"
					}
				}
				default {
					if {$answer != "" && ![regexp $operand $answer]} {
						if {$type != "CMT" || [.fg.$wName current] < 0}	{
							.fg.fgl_$wName configure -foreground red
							append sErrors "\n$name:\t$alert"
						}
					}
				}
			}
		}
	}
	return $sErrors
}
proc signal {s} {
	if {! [winfo exists .condor]} {
		toplevel .condor -relief raised
		set nLines [expr 2 + [string length $s] - [string length [regsub -all "\\n" $s ""]]]
		wm geometry .condor 320x[expr 20*$nLines]
		wm title .condor "FormGen errors"
		pack [label .condor.l -text $s -justify left]
		after 3000 {destroy .condor}
	}
}
proc fg_deMask {s} {
	return [regsub -all "#2(c|C)" [regsub -all "#3(b|B)" $s ";"] ","];	
}
proc fg_Unicode {s} {
	set RE {[][{}\u0100-\uffff]}
	set substitution {[format \\\\u%04x [scan "\\&" %c]]}
	return [subst [regsub -all $RE $s $substitution]]
}
proc mask {s} {
	return [regsub -all "," [regsub -all ";" $s "#3B"] "#2C"]
}
    # variable reMaskJsonChars "\[\\b\\t\\n\\f\\r\\v\\x22\]" ; # BS TAB LF FF CR 11 "
	# variable maskJsonChars [dict create 8 "\\b" 9 "\\t" 10 "\\n" 11 "\\v" 12 "\\f" 13 "\\r" 34 "\\\""]
	# set s [regsub -all "\\x5C" $s {\\\\}]
	# set substitution {[dict get $::Json::maskJsonChars [scan {&} %c]]}
	# return [subst [regsub -all $::Json::reMaskJsonChars $s $substitution]]
	# *** [dict get $::Json::maskJsonChars [scan {
# } %c]]

proc fg_JasonMask {s} {
	return [string map {\x08 {\b} \x09 {\t} \x0a {\n} \x0b {\v} \x0c {\f} \x0d {\r} \x22 {\"} \x2f {\/} \x5c {\\}} $s]
}

proc left {s l} {
	return [string range $s 0 [expr {$l - 1}]]
}
proc mRight {s l} {
	return [string range $s $l [string length $s]]
}
proc fg_File {name} {
	global Elem fields
	set field [lindex $fields [getDictValue $Elem $name ""]]
	lassign $field type varName lbl length extra
	set value [.fg.f[string tolower $varName] get]; # get value from widget if exists
	set file ""
	set folder [pwd]
	if {$value != ""} {
		if {[file isdirectory $value]} {
			set folder $value
		} else {
			set folder [file dirname $value]
			set file [file tail $value]		
		}
	}
	switch $type {
		F {
			set fl [tk_getOpenFile -initialdir $folder -initialfile $file -filetype [fg_genFileTypes $field] -title "Choose a file"]
		}
		WF {
			set fl [tk_getSaveFile -initialdir $folder -initialfile $file -filetype [fg_genFileTypes $field] -title "Save a file"]
		}
		D {
			set fl [tk_chooseDirectory -initialdir $folder -title "Choose a directory"]
		}
	}
	if {$fl != ""} {setValue $name $fl}		
}
proc fg_genFileTypes {field} {
	set l {}
	for {set i 4} {$i<[llength $field]} {incr i} {
		set x [lindex $field $i]
		if {$x != ""} {
			append x ":" $x
			set t [split $x ":"]
			lappend l [list [lindex $t 0] [split [lindex $t 1] " "]]
		}
	}
	return $l
}